/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.interpreters

import java.lang.reflect.Modifier

import scala.collection._
import scala.collection.JavaConversions._

import xgi.ut.dsl.ObjectParsers._
import xgi.ut.dsl.ScenarParsers._
import xgi.ut.dsl.utl.ReflectionWrapper._
import xgi.ut.dsl.utl.ReflectionUtil._
import xgi.ut.dsl.utl.matching._

/**
 * This trait provides the capability of interpreting a ScenarLang script.
 */
trait ScenarioInterpreter extends ToObject  {
  this:{
    val context:Any
    /* 
     * This is where we store the variables being defined in the scenario
     * This field needs to be declared outside the interpreter, that way we
     * can eventually plug-in some values declared at an other place.
     */
    val variables:mutable.Map[String, Any]
  } =>
  
  type ReflectionUtilConfigFn[ResultType, Outcome] = (Any, java.lang.String, java.util.List[MetaMatcher], ResultType) => Outcome
  
  /** interprets one instruction */
  def interpret(i:Instruction):Any = try {
    i match {
      case SyntaxError(msg) => throw new RuntimeException(msg)
      case ConfigInstruction(config) => addConfig(config)
      case Declaration(name, value) => addVariable(name, value)
      case StaticAssignment(stcRef, value) => {
        stcRef.member match {
          case x:FieldRef => injectStaticField(forName(stcRef.declaredType.name), x.name, toObject(value))
          case _ => throw new RuntimeException("It is not possible to assign a result to a static method directly : the class is no mock object")
        }
      }
      case Assignment(destination, value) => interpretComplexInstruction[Any, Object](invokeResult)(destination, value)
      case ExceptionInstruction(destination, value) => interpretComplexInstruction[Throwable, Any](invokeException)(destination, value)
    }
  } catch {
    case ex => {
      throw new RuntimeException(describeError(i, ex), ex) 
    }
  }
  
  private[this] def addVariable(name:String, value:SLiteral) = {
    val valueObject = toObject(value)
    variables += name->valueObject
    valueObject
  }
  
  /** interprets the results, isolates the actual target on which to apply the configFn and applies that function*/
  private[this] def interpretComplexInstruction[ResultType, Outcome](configFn:ReflectionUtilConfigFn[ResultType, Outcome])(destination:MemberRef, value:SLiteral):Any={
    val valueObject = toObject(value).asInstanceOf[ResultType]
    val source = dereference(destination.source, context)
    val interpretation = (source /: destination.invocations.dropRight(1))((newctx, elem) => dereference(elem, newctx))
    configureResult[ResultType, Outcome](configFn)(interpretation, destination.invocations.last, valueObject)
    valueObject
  }
  
  /** assigns a result to a given function invocation or field */
  private[this] def configureResult[ResultType, Outcome](configFn:ReflectionUtilConfigFn[ResultType, Outcome])(target:Any, ref:Reference, result:ResultType = null):Any = ref match {
    case MemberRef(src, destinations) => {
      val source = dereference(src, if(target!=null)target else context)
      val newTarget = (source /: destinations.dropRight(1))((newctx, elem) => dereference(elem, newctx))
      configureResult(configFn)(newTarget, destinations.last, result)
    }
    // an internal would make no sense here
    case FieldRef(name) => {
	  setField(target, name, result)
	  getFieldValue(target, name)
    }
    case FnInvocation(fname, params) => {
      val args = mapParams(target, params)
      configFn(target, fname, args, result)
    }
  }
  
  def mapParams(target:Any, params:Seq[SLiteral]):Seq[MetaMatcher] = params map { p => 
    p match {
      case WildCardObject(Some(declType)) => new MetaAnyMatcher(forName(declType.name))
      case WildCardObject(None) => new MetaAnyMatcher(null)
      case EqMatch(value) => new MetaEqMatcher(toObject(value))
      case DeepEqMatch(value, excl) => new MetaRefEqMatcher(toObject(value), excl)
      case SameMatch(value) => new MetaSameMatcher(toObject(value))
      case ContainsMatch(expr) => new MetaContainsMatcher(expr)
      case MatchesMatch(expr) => new MetaMatchesMatcher(expr)
      case StartsWithMatch(expr) => new MetaStartsWithMatcher(expr)
      case EndsWithMatch(expr) => new MetaEndsWithMatcher(expr)
      case _ =>  new MetaEqMatcher(toObject(p))
    }
  }
  
  def injectStaticField(clazz:Class[_], fieldName:String, value:Any) = {
    val modifiers = getField(null, fieldName)(clazz).getModifiers
    if(Modifier.isFinal(modifiers)) throw new RuntimeException("Sorry, ScenarLang cannot inject the field (%s) because it is declared final".format(fieldName))
    setField(null, fieldName, value)(clazz)
  }
  
  def describeError(instruction:Instruction, ex:Throwable) = 
    """|SCENARLANG INTERPRETATION PROBLEM :
       |Cause : %s
       |Failed (runtime) statement is : 
       |%s
       |""".stripMargin.format(ex.getLocalizedMessage(), instruction.toString)
}
