/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.interpreters

import java.io.File
import java.io.StringReader
import java.util.Collection
import scala.collection.JavaConversions
import scala.collection.JavaConversions.seqAsJavaList
import scala.collection.mutable.Map
import org.mockito.Matchers
import org.mockito.Matchers.any
import org.mockito.Matchers.anyByte
import org.mockito.Matchers.anyChar
import org.mockito.Matchers.anyDouble
import org.mockito.Matchers.anyFloat
import org.mockito.Matchers.anyInt
import org.mockito.Matchers.anyLong
import org.mockito.Matchers.anyShort
import org.mockito.Matchers.anyString
import org.mockito.Mockito.mock
import org.mockito.cglib.proxy.Enhancer
import javax.xml.bind.JAXBContext
import javax.xml.bind.JAXBElement
import javax.xml.transform.stream.StreamSource
import xgi.ut.dsl.ObjectParsers.FieldRef
import xgi.ut.dsl.ObjectParsers.FnInvocation
import xgi.ut.dsl.ObjectParsers.InternalIdentifier
import xgi.ut.dsl.ObjectParsers.JaxbObject
import xgi.ut.dsl.ObjectParsers.MemberRef
import xgi.ut.dsl.ObjectParsers.MockObject
import xgi.ut.dsl.ObjectParsers.Reference
import xgi.ut.dsl.ObjectParsers.Resource
import xgi.ut.dsl.ObjectParsers.SArray
import xgi.ut.dsl.ObjectParsers.SBasicType
import xgi.ut.dsl.ObjectParsers.SCast
import xgi.ut.dsl.ObjectParsers.SConstructor
import xgi.ut.dsl.ObjectParsers.SDeclaredType
import xgi.ut.dsl.ObjectParsers.SLiteral
import xgi.ut.dsl.ObjectParsers.SMap
import xgi.ut.dsl.ObjectParsers.SMatcher
import xgi.ut.dsl.ObjectParsers.SNullObject
import xgi.ut.dsl.ObjectParsers.SProperty
import xgi.ut.dsl.ObjectParsers.SStaticRef
import xgi.ut.dsl.ObjectParsers.SStructure
import xgi.ut.dsl.ObjectParsers.SSyntaxError
import xgi.ut.dsl.ObjectParsers.WildCardObject
import xgi.ut.dsl.utl.ReflectionUtil
import xgi.ut.dsl.utl.ReflectionWrapper.getFieldType
import xgi.ut.dsl.utl.ReflectionWrapper.getFieldValue
import xgi.ut.dsl.utl.ReflectionWrapper.setField

/**
 * This trait gives the mixed class/object the ability to let
 * a JsonLiteral enter the SStructured+typed world of Java & Scala
 */
trait ToObject extends ConversionConfigInterpreter {
  this:{val context:Any; val variables:Map[String, Any]} =>
  /**
   * Let a Literal enter the SStructured+typed world of Java & Scala.
   * A basic type becomes an instance of this basic type.
   * An SArray becomes a java SArray
   * A complextype is transformed into a map where each of the fields is a key.
   */
  def toObject(json:SLiteral):Any = json match {
    // syntax error
    case SSyntaxError(mesg) => Console.err.println(mesg); mesg
    
    case SCast(SDeclaredType(dclType), value) => cast(toObject(value), forName(dclType))
    
    // basic types
    case SBasicType(value, None) => value
    case SBasicType(text:String,  Some(SDeclaredType(typeName))) => convert(text.toString, forName(typeName))
    case SBasicType(num:Number,  Some(SDeclaredType(typeName))) =>  ReflectionUtil.autoBox(num, forName(typeName))
    
    // SArray
    case SArray(elems, optionalType) => {
      val items = elems map toObject
      optionalType match {
        case None => items.toArray
        case Some(SDeclaredType(typeName)) => castCollection(items, typeName)
      }
    }
    
    // SMap
    case SMap(associations) => {
      val ret = new java.util.HashMap[Any, Any]
      associations foreach {entry => ret.put(toObject(entry.key), toObject(entry.value))}
      ret
    }
    
    // Special objects
    case SNullObject => null
    case MockObject(SDeclaredType(typeName)) => mock(forName(typeName))
    case Resource(location) => new File(getClass().getResource(location).getFile())
    case JaxbObject(resource, SDeclaredType(typeName)) => {
      val file = toObject(resource).asInstanceOf[File]
      val xml = ("" /: io.Source.fromFile(file).getLines)(_+_)
      fromJaxb(xml, forName(typeName))
    }
   
    // Reference
    case x:Reference => dereference(x, context)
    // Matchers
    case x:SMatcher => throw new RuntimeException("Wrong usage of the matcher : they sould only be used as last step in configuration of a mock object")
    
    // SStructure types
    case SStructure(props, None, _) => {
      var theMap = Map.empty[String, Any]
      props map {p => theMap+= p.key -> toObject(p.value)}
      theMap
    }
    
    case SStructure(props, Some(SDeclaredType(typeName)), constructor) => {
      val clazz = forName(typeName)
      val instance = instanciate(clazz, constructor)
      props foreach { x => 
        val value = toObject(x.value) match {
          case y:Map[String, Any] => {
            val theClass = getFieldType(instance, x.key)
            // #1 - (attempt) make it possible to inject a map field
            if(ReflectionUtil.isAssignableFrom(theClass, y.getClass())) {
              y
            } else if(ReflectionUtil.isAssignableFrom(theClass, classOf[java.util.Map[_,_]])) {
              JavaConversions.mutableMapAsJavaMap(y)
            } else {
              val valueInstance = instanciate(theClass)
              fromMap(valueInstance, y)
            }
          }
          case y:String => convert(y, getFieldType(instance, x.key))
          case other => other
        }
        setField(instance, x.key, value)
       }
      instance
    } 
    case other => other
  }
  
  def cast(value:Any, to:Class[_]) = value match {
    case x:Number => ReflectionUtil.autoBox(value, to)
    case _ => to.cast(value) 
  }  
  
  /** dereferences a reference */
  def dereference(ref:Reference, currentContext:Any):Any = ref match {
    case InternalIdentifier(name) => variables(name)
    case SStaticRef(declType, member) => staticDereference(member, forName(declType.name))
    case MemberRef(source, invocations) => {
      val src = dereference(source, currentContext)
      (src /: invocations)((ctx, elem) => dereference(elem, ctx))
    }
    case FieldRef(name) => {
      var externalValue = getFieldValue(currentContext, name)
      // avoid NPE
      if(externalValue==null){
        externalValue = mock(getFieldType(currentContext, name))
        setField(currentContext, name, externalValue)
      }
      externalValue
    }
    case FnInvocation(fname, params) => {
      if(! params.forall {!_.isInstanceOf[SMatcher]}) throw new RuntimeException("Usage of matchers/wildcards is not allowed for method traversals.\n" +
      		"They can only be used as last step to a result assignment");
      val args = params map {toObject}
      ReflectionUtil.invoke(currentContext, fname, args)
    }
  }
  
  /** dereferences a static ref */
  def staticDereference(ref:Reference, clazz:Class[_]):Any = ref match {
    case FieldRef(name) => getFieldValue(null, name)(clazz)
    case FnInvocation(name, args) => ReflectionUtil.invokeStatic(clazz, name, args map {toObject})
    case MemberRef(source, invocations) => {
      val src = staticDereference(source, clazz)
      (src /: invocations)((ctx, elem) => dereference(elem, ctx))
    }
  }
  
  /** 
   * this will instanciate an object of the class 'clazz' using its default constructor
   * even if it is private.
   */
  def instanciate[A](clazz:Class[A], sconstr:Option[SConstructor]=None):A={
    val args = sconstr match {
      case Some(SConstructor(params)) => params map {toObject}
      case None => List.empty
    }
    ReflectionUtil.instanciate(clazz , args).asInstanceOf[A]
  }
  
  /** injects all the properties in the given object */
  private[this] def fromMap(target:Any, values:Map[String, Any]):Any = {
    values foreach {y => setField(target, y._1, y._2)}
    target
  }
  
  /** unmarshals a jaxb object from a given resource */
  private[this] def fromJaxb(xml:String, clazz:Class[_]):Any = {
    val jaxb = JAXBContext.newInstance(clazz)
    val unmarsh = jaxb.createUnmarshaller()
    val unmarshalled = unmarsh.unmarshal(new StreamSource(new StringReader(xml)), clazz)
    unmarshalled match {
      case x:JAXBElement[_] => x.getValue()
      case _ => unmarshalled
    }
  }
 
  private[this] def castCollection(items:Seq[_], to:String) = {
    val collectionType = forName(to)
    if(!classOf[Collection[_]].isAssignableFrom(collectionType)){ 
      throw new RuntimeException(to+" is not an implementation of java.util.Collection");
    } else {
      val retVal = instanciate(collectionType).asInstanceOf[Collection[Any]]
      val collection:java.util.List[Any] = items.toList
      retVal.addAll(collection)
      retVal
    }
  }
}
